import sys
import base64
from base64 import b64encode, b64decode
import traceback
import os

try:
    from StringIO import StringIO
except:
    from io import BytesIO

from PIL import Image
from six import string_types


def pdf_preprocess(pdf, batch=False):
    """
    Load pdfs from local filepath if not already b64 encoded
    """
    if batch:
        return [pdf_preprocess(doc, batch=False) for doc in pdf]

    if os.path.isfile(pdf):
        # a filepath is provided, read and encode
        return b64encode(open(pdf, "rb").read())
    else:
        # assume pdf is already b64 encoded
        return pdf


def postprocess_image(image):
    raw_data = image.get("data")

    try:
        if (2, 6) <= sys.version_info < (3, 0):
            data = b64decode(raw_data)
            return Image.open(StringIO(data))
        elif sys.version_info >= (3, 0):
            if not isinstance(raw_data, bytes):
                raw_data = bytes(raw_data, "utf-8")
            data = base64.decodestring(raw_data)
            return Image.open(BytesIO(data))
        else:
            raise AssertionError(
                "Unsupport python version: {version}".format(
                    version=str(sys.version_info)
                )
            )
    except IOError:
        traceback.print_exc()
        return None


def postprocess_images(images):
    images = [postprocess_image(image) for image in images]
    images = [image for image in images if image]  # remove Nones
    return images



"""
Image Utils
Handles preprocessing images before they are sent to the server
"""
import os.path, base64, re, warnings
from six import BytesIO, string_types, PY3

from PIL import Image

from indicoio.utils.errors import IndicoError

B64_PATTERN = re.compile(
    "^([A-Za-z0-9+/]{4})*([A-Za-z0-9+/]{4}|[A-Za-z0-9+/]{3}=|[A-Za-z0-9+/]{2}==)"
)


def file_exists(filename):
    """
    Check if a file exists (and don't error out on unicode inputs)
    """
    try:
        return os.path.isfile(filename)
    except (UnicodeDecodeError, UnicodeEncodeError, ValueError):
        return False


def data_preprocess(data, size=None, min_axis=None, batch=False):
    """
    Takes data and prepares it for sending to the api including
    resizing and image data/structure standardizing.
    """
    if batch:
        return [
            data_preprocess(el, size=size, min_axis=min_axis, batch=False)
            for el in data
        ]

    if isinstance(data, dict):
        return {
            key: data_preprocess(value, size=size, min_axis=min_axis, batch=False)
            for key, value in data.items()
        }
    if isinstance(data, string_types):
        if file_exists(data):
            # probably a path to an image
            preprocessed = Image.open(data)
        else:
            # base 64 encoded image data, a url, or raw content
            # send raw data to the server and let the server infer type
            b64_or_url = re.sub("^data:image/.+;base64,", "", data)
            return b64_or_url

    elif isinstance(data, Image.Image):
        # data is image from PIL
        preprocessed = data

    elif type(data).__name__ == "ndarray":
        # data is likely image from numpy/scipy
        if "float" in str(data.dtype) and data.min() >= 0 and data.max() <= 1:
            data *= 255.0
        try:
            preprocessed = Image.fromarray(data.astype("uint8"))
        except TypeError:
            raise IndicoError(
                "Please ensure the numpy array is in a format by PIL. "
                "Values must be between 0 and 1 or between 0 and 255 in greyscale, rgb, or rgba format."
            )

    else:
        # at this point we are unsure of the type -- it could be malformatted text or image data.
        raise IndicoError(
            "Invalid input datatype: `{}`. "
            "Ensure input data is one of the following types: "
            "`str`, `unicode`, `PIL.Image`, `np.ndarray`.".format(
                data.__class__.__name__
            )
        )

    #
    if size or min_axis:
        preprocessed = resize_image(preprocessed, size, min_axis)

    # standardize on b64 encoding for sending image data over the wire
    temp_output = BytesIO()
    preprocessed.save(temp_output, format="PNG")
    temp_output.seek(0)
    output_s = temp_output.read()
    return (
        base64.b64encode(output_s).decode("utf-8")
        if PY3
        else base64.b64encode(output_s)
    )


def resize_image(image, size, min_axis):
    if min_axis:
        min_idx, other_idx = (0, 1) if image.size[0] < image.size[1] else (1, 0)
        aspect = image.size[other_idx] / float(image.size[min_idx])
        if aspect > 10:
            warnings.warn(
                "An aspect ratio greater than 10:1 is not recommended", Warning
            )
        size_arr = [0, 0]
        size_arr[min_idx] = size
        size_arr[other_idx] = int(size * aspect)
        image = image.resize(tuple(size_arr))
    elif size:
        image = image.resize(size)
    return image


def get_list_dimensions(_list):
    """
    Takes a nested list and returns the size of each dimension followed
    by the element type in the list
    """
    if isinstance(_list, list) or isinstance(_list, tuple):
        return [len(_list)] + get_list_dimensions(_list[0])
    return []


def get_element_type(_list, dimens):
    """
    Given the dimensions of a nested list and the list, returns the type of the
    elements in the inner list.
    """
    elem = _list
    for _ in range(len(dimens)):
        elem = elem[0]
    return type(elem)



from ..utils.api import api_handler
from ..utils.decorators import detect_batch_decorator


@detect_batch_decorator
def summarization(text, cloud=None, batch=False, api_key=None, version=1, **kwargs):
    """
    Given input text, returns a `top_n` length sentence summary.

    Example usage:

    .. code-block:: python

       >>> from indicoio import summarization
       >>> summary = summarization("https://en.wikipedia.org/wiki/Yahoo!_data_breach")
       >>> summary
       ["This information was disclosed two years later on September 22, 2016.", "[1] The data breach is one of the largest in the history of the Internet.", "Specific details of material taken include names, email addresses, telephone numbers, dates of birth, and encrypted passwords.", "[2]\\n\\nEvents [ edit ]\\n\\nYahoo alleged in its statement that the breach was carried out by \\"state-sponsored\\" hackers,[3] but the organization did not name any country.", "We had our own use for it and other buyers did as well."]

    :param text: The text to be analyzed.
    :type text: str or unicode
    :rtype: Dictionary of party probability pairs
    """
    url_params = {"batch": batch, "api_key": api_key, "version": version}
    return api_handler(
        text, cloud=cloud, api="summarization", url_params=url_params, **kwargs
    )



from .content_filtering import content_filtering
from .facial_features import facial_features
from .facial_localization import facial_localization
from .fer import fer
from .image_features import image_features
from .image_recognition import image_recognition

IMAGE_APIS = {
    "content_filtering": content_filtering,
    "facial_features": facial_features,
    "facial_localization": facial_localization,
    "fer": fer,
    "image_features": image_features,
    "image_recognition": image_recognition,
}



from .utils import multi, MULTIAPI_NOT_SUPPORTED
from ..text import TEXT_APIS
from ..utils.decorators import detect_batch_decorator

DEFAULT_APIS = list(set(TEXT_APIS.keys()) - set(MULTIAPI_NOT_SUPPORTED))


@detect_batch_decorator
def analyze_text(input_text, apis=DEFAULT_APIS, **kwargs):
    """
    Given input text, returns the results of specified text apis. Possible apis
    include: [ 'text_tags', 'political', 'sentiment', 'language' ]

    Example usage:

    .. code-block:: python

       >>> import indicoio
       >>> text = 'Monday: Delightful with mostly sunny skies. Highs in the low 70s.'
       >>> results = indicoio.analyze_text(data = text, apis = ["language", "sentiment"])
       >>> language_results = results["language"]
       >>> sentiment_results = results["sentiment"]

    :param text: The text to be analyzed.
    :param apis: List of apis to use.
    :type text: str or unicode
    :type apis: list of str
    :rtype: Dictionary of api responses
    """
    cloud = kwargs.pop("cloud", None)
    batch = kwargs.pop("batch", False)
    api_key = kwargs.pop("api_key", None)

    return multi(
        data=input_text,
        datatype="text",
        cloud=cloud,
        batch=batch,
        api_key=api_key,
        apis=apis,
        accepted_apis=TEXT_APIS,
        **kwargs
    )



from ..utils.api import api_handler
from ..utils.decorators import detect_batch_decorator


@detect_batch_decorator
def text_features(text, cloud=None, batch=False, api_key=None, version=None, **kwargs):
    """
    Given input text, returns a numeric feature vector that represents the content.

    Example usage:

    .. code-block:: python

       >>> from indicoio import text_features
       >>> text_features("Queen of England")
       [0.04509247093572533, -0.052756784338865576, ...]

    :param text: The text to be analyzed.
    :type text: str or unicode
    :rtype: List of floats which represents the content of the input text
    """
    url_params = {"batch": batch, "api_key": api_key, "version": version}
    kwargs["synonyms"] = False
    return api_handler(
        text, cloud=cloud, api="textfeatures", url_params=url_params, **kwargs
    )



from importlib import import_module

from .language import language
from .personality import personality
from .political import political
from .summarization import summarization
from .emotion import emotion
from .organizations import organizations
from .personas import personas
from .relevance import relevance
from .text_features import text_features
from .keywords import keywords
from .people import people
from .places import places
from .sentiment import sentiment
from .sentiment_hq import sentiment_hq
from .text_tags import text_tags

TEXT_APIS = {
    "language": language,
    "personality": personality,
    "political": political,
    "summarization": summarization,
    "emotion": emotion,
    "organizations": organizations,
    "personas": personas,
    "relevance": relevance,
    "text_features": text_features,
    "people": people,
    "places": places,
    "sentiment": sentiment,
    "sentiment_hq": sentiment_hq,
    "text_tags": text_tags,
}



import os
import sys

try:
    import configparser
except ImportError:
    import ConfigParser as configparser


PY2 = True if sys.version_info[0] == 2 else False
PY3 = not PY2


class Settings(configparser.ConfigParser):
    def __init__(self, *args, **kwargs):
        """
        files: filepaths or open file objects
        """
        self.files = kwargs.pop("files")

        configparser.ConfigParser.__init__(self, *args, **kwargs)

        for fd in self.files:
            try:
                self.read_file(fd)
            except AttributeError:
                # python 2
                try:
                    self.readfp(fd)
                except AttributeError:
                    self.read(fd)
            except configparser.MissingSectionHeaderError:
                self.read(fd)

        self.auth_settings = self.get_section("auth")
        self.private_cloud_settings = self.get_section("private_cloud")

    def get_section(self, section):
        """
        Retrieve a configparser section as a dictionary, default to {}
        """
        try:
            return dict(self.items(section))
        except configparser.NoSectionError:
            return {}

    def cloud(self):
        return (
            os.getenv("INDICO_CLOUD")
            or self.private_cloud_settings.get("cloud")
            or None
        )

    def api_key(self):
        return os.getenv("INDICO_API_KEY") or self.auth_settings.get("api_key") or None


SETTINGS = Settings(
    files=[os.path.expanduser("~/.indicorc"), os.path.join(os.getcwd(), ".indicorc")]
)

api_key = SETTINGS.api_key()
cloud = SETTINGS.cloud()
host = os.getenv("INDICO_API_HOST", "apiv2.indico.io")
url_protocol = "https"
serializer = "msgpack"



from ..utils.docx import docx_preprocess
from ..utils.api import api_handler
from ..utils.decorators import detect_batch_decorator


@detect_batch_decorator
def docx_extraction(
    docx, cloud=None, batch=False, api_key=None, version=None, **kwargs
):
    """
    Given a .docx file, returns the raw text associated with the given .docx file.
    The .docx file may be provided as base64 encoded data or as a filepath.

    Example usage:

    .. code-block:: python

       >>> from indicoio import docx_extraction
       >>> results = docx_extraction(docx_file)

    :param docx: The docx to be analyzed.
    :type docx: str or list of strs
    :rtype: dict or list of dicts
    """
    docx = docx_preprocess(docx, batch=batch)
    url_params = {"batch": batch, "api_key": api_key, "version": version}
    results = api_handler(
        docx, cloud=cloud, api="docxextraction", url_params=url_params, **kwargs
    )
    return results



from ..utils.preprocessing import data_preprocess
from ..utils.api import api_handler
from ..utils.decorators import detect_batch_decorator


@detect_batch_decorator
def facial_localization(
    image, cloud=None, batch=False, api_key=None, version=None, **kwargs
):
    """
    Given an image, returns a list of faces found within the image.
    For each face, we return a dictionary containing the upper left corner and lower right corner.
    If crop is True, the cropped face is included in the dictionary.
    Input should be in a numpy ndarray or a filename.

    Example usage:

    .. code-block:: python

       >>> from indicoio import facial_localization
       >>> import numpy as np
       >>> img = np.zeros([image of a face])
       >>> faces = facial_localization(img)
       >>> len(faces)
       1

    :param image: The image to be analyzed.
    :type image: filepath or ndarray
    :rtype: List of faces (dict) found.
    """
    image = data_preprocess(image, batch=batch)
    url_params = {"batch": batch, "api_key": api_key, "version": version}
    return api_handler(
        image, cloud=cloud, api="faciallocalization", url_params=url_params, **kwargs
    )



from ..utils.api import api_handler
from ..utils.decorators import detect_batch_decorator


@detect_batch_decorator
def personality(text, cloud=None, batch=False, api_key=None, version=None, **kwargs):
    """
    Given input text, returns the authors 'Extraversion', 'Conscientiousness',
    'Openness', and 'Agreeableness' score (a float between 0 and 1) in a dictionary.

    Example usage:

    .. code-block:: python

       >>> text = "I love going out with my friends"
       >>> entities = indicoio.personality(text)
       {'Extraversion': 0.69691890478134155, 'Conscientiousness': 0.4658474326133728,
        'Openness': 0.42654544115066528, 'Agreeableness': 0.7414245903}

    :param text: The text to be analyzed.
    :type text: str or unicode
    :rtype: The authors 'Extraversion', 'Conscientiousness',
    'Openness', and 'Agreeableness' score (a float between 0 and 1) in a dictionary.
    """
    url_params = {"batch": batch, "api_key": api_key, "version": version}
    return api_handler(
        text, cloud=cloud, api="personality", url_params=url_params, **kwargs
    )



from ..utils.preprocessing import data_preprocess
from ..utils.api import api_handler
from ..utils.decorators import detect_batch_decorator


@detect_batch_decorator
def facial_features(
    image, cloud=None, batch=False, api_key=None, version=None, **kwargs
):
    """
    Given an grayscale input image of a face, returns a 48 dimensional feature vector explaining that face.
    Useful as a form of feature engineering for face oriented tasks.
    Input should be in a list of list format, resizing will be attempted internally but for best
    performance, images should be already sized at 48x48 pixels.

    Example usage:

    .. code-block:: python

       >>> from indicoio import facial_features
       >>> import numpy as np
       >>> face = np.zeros((48,48))
       >>> features = facial_features(face)
       >>> len(features)
       48

    :param image: The image to be analyzed.
    :type image: list of lists
    :rtype: List containing feature responses
    """
    image = data_preprocess(
        image, batch=batch, size=None if kwargs.get("detect") else (48, 48)
    )
    url_params = {"batch": batch, "api_key": api_key, "version": version}
    return api_handler(
        image, cloud=cloud, api="facialfeatures", url_params=url_params, **kwargs
    )



from ..utils.api import api_handler
from ..utils.decorators import detect_batch_decorator


@detect_batch_decorator
def organizations(text, cloud=None, batch=None, api_key=None, version=2, **kwargs):
    """
    Given input text, returns references to specific organizations found in the text

    Example usage:

    .. code-block:: python

       >>> text = "London Underground's boss Mike Brown warned that the strike ..."
       >>> entities = indicoio.organizations(text)
        [
          {
            u'text': "London Underground",
            u'confidence': 0.8643872141838074,
            u'position': [0, 18]
          }
        ]

    :param text: The text to be analyzed.
    :type text: str or unicode
    :rtype: Dictionary of language probability pairs
    """
    url_params = {"batch": batch, "api_key": api_key, "version": version}
    return api_handler(
        text, cloud=cloud, api="organizations", url_params=url_params, **kwargs
    )



from ..utils.api import api_handler
from ..utils.decorators import detect_batch_decorator


@detect_batch_decorator
def relevance(
    data, queries, cloud=None, batch=False, api_key=None, version=None, **kwargs
):
    """
    Given input text and a list of query terms / phrases, returns how relevant the query is
    to the input text.

    Example usage:

    .. code-block:: python

       >>> import indicoio
       >>> text = 'On Monday, president Barack Obama will be giving his keynote address at...'
       >>> relevance = indicoio.relevance(text, queries=['president'])
       >>> print "Relevance: " + str(relevance[0])
       u'Relevance: [0.44755361996336784]'

    :param text: The text to be analyzed.
    :param queries: a list of terms or phrases to measure similarity against
    :type text: str or unicode
    :rtype: Dictionary of feature score pairs
    """
    url_params = {"batch": batch, "api_key": api_key, "version": version}
    kwargs["queries"] = queries
    kwargs["synonyms"] = False
    return api_handler(
        data, cloud=cloud, api="relevance", url_params=url_params, **kwargs
    )



from ..utils.api import api_handler
from ..utils.decorators import detect_batch_decorator


@detect_batch_decorator
def keywords(
    text, cloud=None, batch=False, api_key=None, version=2, batch_size=None, **kwargs
):
    """
    Given input text, returns series of keywords and associated scores

    Example usage:

    .. code-block:: python

       >>> import indicoio
       >>> import numpy as np
       >>> text = 'Monday: Delightful with mostly sunny skies. Highs in the low 70s.'
       >>> keywords = indicoio.keywords(text, top_n=3)
       >>> print "The keywords are: "+str(keywords.keys())
       u'The keywords are ['delightful', 'highs', 'skies']

    :param text: The text to be analyzed.
    :type text: str or unicode
    :rtype: Dictionary of feature score pairs
    """
    if kwargs.get("language", "english") != "english":
        version = 1
    url_params = {"batch": batch, "api_key": api_key, "version": version}
    return api_handler(
        text,
        cloud=cloud,
        api="keywords",
        url_params=url_params,
        batch_size=batch_size,
        **kwargs
    )



from .utils import multi, MULTIAPI_NOT_SUPPORTED
from ..image import IMAGE_APIS
from ..utils.preprocessing import data_preprocess
from ..utils.decorators import detect_batch_decorator

DEFAULT_APIS = list(set(IMAGE_APIS.keys()) - set(MULTIAPI_NOT_SUPPORTED))


@detect_batch_decorator
def analyze_image(image, apis=DEFAULT_APIS, **kwargs):
    """
    Given input image, returns the results of specified image apis. Possible apis
    include: ['fer', 'facial_features', 'image_features']

    Example usage:

    .. code-block:: python

       >>> import indicoio
       >>> import numpy as np
       >>> face = np.zeros((48,48)).tolist()
       >>> results = indicoio.analyze_image(image = face, apis = ["fer", "facial_features"])
       >>> fer = results["fer"]
       >>> facial_features = results["facial_features"]

    :param text: The text to be analyzed.
    :param apis: List of apis to use.
    :type text: str or unicode
    :type apis: list of str
    :rtype: Dictionary of api responses
    """
    cloud = kwargs.pop("cloud", None)
    batch = kwargs.pop("batch", False)
    api_key = kwargs.pop("api_key", None)

    return multi(
        data=data_preprocess(image, batch=batch),
        datatype="image",
        cloud=cloud,
        batch=batch,
        api_key=api_key,
        apis=apis,
        accepted_apis=IMAGE_APIS,
        **kwargs
    )



from functools import wraps


def detect_batch(data):
    return isinstance(data, list)


def detect_batch_decorator(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        if isinstance(args[0], list):
            kwargs["batch"] = True
        return f(*args, **kwargs)

    return wrapper



import concurrent.futures
from concurrent.futures import ThreadPoolExecutor, as_completed

from ..utils.errors import IndicoError
from ..utils.decorators import detect_batch_decorator

MULTIAPI_NOT_SUPPORTED = ["relevance"]

EXECUTOR = ThreadPoolExecutor(max_workers=4)


def multi(data, datatype, apis, accepted_apis, batch=False, **kwargs):
    """
    Helper to make multi requests of different types.

    :param data: Data to be sent in API request
    :param datatype: String type of API request
    :param apis: List of apis to use.
    :param batch: Is this a batch request?
    :rtype: Dictionary of api responses
    """
    # Client side api name checking - strictly only accept func name api
    invalid_apis = [
        api for api in apis if api not in accepted_apis or api in MULTIAPI_NOT_SUPPORTED
    ]
    if invalid_apis:
        raise IndicoError(
            "The following are not valid %s APIs: %s. Please reference the available APIs below:\n%s"
            % (datatype, ", ".join(invalid_apis), ", ".join(accepted_apis.keys()))
        )

    # Convert client api names to server names before sending request
    cloud = kwargs.pop("cloud", None)
    api_key = kwargs.pop("api_key", None)

    api_results_executor = {}
    for api in apis:
        api_results_executor[
            EXECUTOR.submit(
                accepted_apis[api],
                data,
                cloud=cloud,
                api_key=api_key,
                batch=batch,
                **kwargs
            )
        ] = api

    api_results = {}
    for future in concurrent.futures.as_completed(api_results_executor):
        api_results[api_results_executor[future]] = future.result()

    return api_results



import time
import sys

from ..utils.api import api_handler
from ..utils.decorators import detect_batch
from ..utils.preprocessing import data_preprocess
from ..utils.errors import IndicoError


def _unpack_list(example):
    """
    Input data format standardization
    """
    try:
        x = example[0]
        y = example[1]
        meta = None
        return x, y, meta
    except IndexError:
        raise IndicoError(
            "Invalid input data.  Please ensure input data is "
            "formatted as a list of `[data, target]` pairs."
        )


def _unpack_dict(example, target=None):
    """
    Input data format standardization
    """
    if target is None:
        try:
            x = example["data"]
            y = example["target"]
            meta = example.get("metadata", {})
            return x, y, meta
        except KeyError:
            raise IndicoError(
                "Invalid input data.  Please ensure input data is "
                "formatted as a list of dicts with `data` and `target` keys. "
                "A `metadata` key may optionally be included."
            )
    else:
        # multi-field input
        y = example.pop(target, None)
        x = example
        meta = {}
        return x, y, meta


def _unpack_data(data, target=None):
    """
    Break Xs, Ys, and metadata out into separate lists for data preprocessing.
    Run basic data validation.
    """
    xs = [None] * len(data)
    ys = [None] * len(data)
    metadata = [None] * len(data)
    for idx, example in enumerate(data):
        if isinstance(example, (list, tuple)):
            xs[idx], ys[idx], metadata[idx] = _unpack_list(example)
        if isinstance(example, dict):
            xs[idx], ys[idx], metadata[idx] = _unpack_dict(example, target=target)

    return xs, ys, metadata


def _pack_data(X, Y, metadata, target=None):
    """
    After modifying / preprocessing inputs,
    reformat the data in preparation for JSON serialization
    """
    if target is None:
        if not any(metadata):
            # legacy list of list format is acceptable
            return list(zip(X, Y))

        else:
            # newer dictionary-based format is required in order to save metadata
            return [
                {"data": x, "target": y, "metadata": meta}
                for x, y, meta in zip(X, Y, metadata)
            ]
    else:
        # multi-field input
        for i in range(len(X)):
            assert isinstance(X[i], dict)
            X[i][target] = Y[i]
        return X


def visualize_explanation(explanation, label=None):
    """
    Given the output of the explain() endpoint, produces a terminal visual that plots response strength over a sequence
    """
    if not sys.version_info[:2] >= (3, 5):
        raise IndicoError("Python >= 3.5+ is required for explanation visualization")

    try:
        from colr import Colr as C
    except ImportError:
        raise IndicoError(
            "Package colr >= 0.8.1 is required for explanation visualization."
        )

    cursor = 0
    text = explanation["text"]
    for token in explanation.get("token_predictions"):
        try:
            class_confidence = token.get("prediction")[label]
        except KeyError:
            raise IndicoError("Invalid label: {}".format(label))

        if class_confidence > 0.5:
            fg_color = (255, 255, 255)
        else:
            fg_color = (0, 0, 0)
        rg_value = 255 - int(class_confidence * 255)
        token_end = token.get("token").get("end")
        token_text = text[cursor:token_end]
        cursor = token_end
        sys.stdout.write(
            str(
                C()
                .b_rgb(rg_value, rg_value, 255)
                .rgb(fg_color[0], fg_color[1], fg_color[2], token_text)
            )
        )
    sys.stdout.write("\n")
    sys.stdout.flush()


class Collection(object):
    def __init__(self, collection, *args, **kwargs):
        self.keywords = {
            "domain": kwargs.get("domain"),
            "shared": kwargs.get("shared"),
            "collection": collection,
        }

    def _api_handler(self, *args, **kwargs):
        """
        Thin wrapper around api_handler from `indicoio.utils.api` to add in stored keyword argument to the JSON body
        """
        keyword_arguments = {}
        keyword_arguments.update(self.keywords)
        keyword_arguments.update(kwargs)
        return api_handler(*args, **keyword_arguments)

    def add_data(
        self, data, cloud=None, batch=False, api_key=None, version=None, **kwargs
    ):
        """
        This is the basic training endpoint. Given a piece of text and a score, either categorical
        or numeric, this endpoint will train a new model given the additional piece of information.

        Inputs
        data - List: The text and collection/score associated with it. The length of the text (string) should ideally
          be longer than 100 characters and contain at least 10 words. While the API will support
          shorter text, you will find that the accuracy of results improves significantly with longer
          examples. For an additional fee, this end point will support image input as well. The collection/score
          can be a string or float. This is the variable associated with the text. This can either be categorical
          (the tag associated with the post) or numeric (the number of Facebook shares the post
          received). However it can only be one or another within a given label.
        domain (optional) - String: This is an identifier that helps determine the appropriate techniques for indico
          to use behind the scenes to train your model.  One of {"standard", "topics"}.
        api_key (optional) - String: Your API key, required only if the key has not been declared
          elsewhere. This allows the API to recognize a request as yours and automatically route it
          to the appropriate destination.
        cloud (optional) - String: Your private cloud domain, required only if the key has not been declared
          elsewhere. This allows the API to recognize a request as yours and automatically route it
          to the appropriate destination.
        """
        if not len(data):
            raise IndicoError("No input data provided.")
        batch = isinstance(data[0], (list, tuple, dict))

        # standarize format for preprocessing batch of examples
        if not batch:
            data = [data]

        target = kwargs.get("target")
        X, Y, metadata = _unpack_data(data, target=target)
        X = data_preprocess(X, batch=True)
        data = _pack_data(X, Y, metadata, target=target)

        # if a single example was passed in, unpack
        if not batch:
            data = data[0]

        url_params = {
            "batch": batch,
            "api_key": api_key,
            "version": version,
            "method": "add_data",
        }
        return self._api_handler(
            data, cloud=cloud, api="custom", url_params=url_params, **kwargs
        )

    def train(self, cloud=None, batch=False, api_key=None, version=None, **kwargs):
        """
        This is the basic training endpoint. Given an existing dataset this endpoint will train a model.

        Inputs
        api_key (optional) - String: Your API key, required only if the key has not been declared
          elsewhere. This allows the API to recognize a request as yours and automatically route it
          to the appropriate destination.
        cloud (optional) - String: Your private cloud domain, required only if the key has not been declared
          elsewhere. This allows the API to recognize a request as yours and automatically route it
          to the appropriate destination.
        """
        url_params = {
            "batch": batch,
            "api_key": api_key,
            "version": version,
            "method": "train",
        }
        return self._api_handler(
            self.keywords["collection"],
            cloud=cloud,
            api="custom",
            url_params=url_params,
            **kwargs
        )

    def predict(
        self, data, cloud=None, batch=False, api_key=None, version=None, **kwargs
    ):
        """
        This is the prediction endpoint. This will be the primary interaction point for all predictive
        analysis.

        Inputs
        data - String: The text example being provided to the API. As a general rule, the data should be as
          similar to the examples given to the train function (above) as possible. Because language
          in different domains is used very differently the accuracy will generally drop as the
          difference between this text and the training text increases. Base64 encoded image data, image urls, and
          text content are all valid.
        domain (optional) - String: This is an identifier that helps determine the appropriate techniques for indico
          to use behind the scenes to train your model.  One of {"standard", "topics"}.
        api_key (optional) - String: Your API key, required only if the key has not been declared
          elsewhere. This allows the API to recognize a request as yours and automatically route it
          to the appropriate destination.
        cloud (optional) - String: Your private cloud domain, required only if the key has not been declared
          elsewhere. This allows the API to recognize a request as yours and automatically route it
          to the appropriate destination.
        """
        batch = detect_batch(data)
        data = data_preprocess(data, batch=batch)
        url_params = {"batch": batch, "api_key": api_key, "version": version}
        return self._api_handler(
            data, cloud=cloud, api="custom", url_params=url_params, **kwargs
        )

    def explain(
        self, data, cloud=None, batch=False, api_key=None, version=None, **kwargs
    ):
        """
        This is the explain endpoint. This allows for predictions that also include information
        about the training data that led to the models decision.

        Inputs
        data - String: The text example being provided to the API. As a general rule, the data should be as
          similar to the examples given to the train function (above) as possible. Because language
          in different domains is used very differently the accuracy will generally drop as the
          difference between this text and the training text increases. Base64 encoded image data, image urls, and
          text content are all valid.
        domain (optional) - String: This is an identifier that helps determine the appropriate techniques for indico
          to use behind the scenes to train your model.  One of {"standard", "topics"}.
        api_key (optional) - String: Your API key, required only if the key has not been declared
          elsewhere. This allows the API to recognize a request as yours and automatically route it
          to the appropriate destination.
        cloud (optional) - String: Your private cloud domain, required only if the key has not been declared
          elsewhere. This allows the API to recognize a request as yours and automatically route it
          to the appropriate destination.
        """
        batch = detect_batch(data)
        data = data_preprocess(data, batch=batch)
        url_params = {
            "batch": batch,
            "api_key": api_key,
            "version": version,
            "method": "explain",
        }
        return self._api_handler(
            data, cloud=cloud, api="custom", url_params=url_params, **kwargs
        )

    def clear(self, cloud=None, api_key=None, version=None, **kwargs):
        """
        This is an API made to remove all of the data associated from a given colletion. If there's been a data
        corruption issue, or a large amount of incorrect data has been fed into the API it is often difficult
        to correct. This allows you to clear a colletion and start from scratch. Use with caution! This is not
        reversible.

        Inputs
        api_key (optional) - String: Your API key, required only if the key has not been declared
          elsewhere. This allows the API to recognize a request as yours and automatically route it
          to the appropriate destination.
        cloud (optional) - String: Your private cloud domain, required only if the key has not been declared
          elsewhere. This allows the API to recognize a request as yours and automatically route it
          to the appropriate destination.
        """
        url_params = {
            "batch": False,
            "api_key": api_key,
            "version": version,
            "method": "clear_collection",
        }
        return self._api_handler(
            None, cloud=cloud, api="custom", url_params=url_params, **kwargs
        )

    def info(self, cloud=None, api_key=None, version=None, **kwargs):
        """
        Return the current state of the model associated with a given collection
        """
        url_params = {
            "batch": False,
            "api_key": api_key,
            "version": version,
            "method": "info",
        }
        return self._api_handler(
            None, cloud=cloud, api="custom", url_params=url_params, **kwargs
        )

    def remove_example(
        self, data, cloud=None, batch=False, api_key=None, version=None, **kwargs
    ):
        """
        This is an API made to remove a single instance of training data. This is useful in cases where a
        single instance of content has been modified, but the remaining examples remain valid. For
        example, if a piece of content has been retagged.

        Inputs
        data - String: The exact text you wish to remove from the given collection. If the string
          provided does not match a known piece of text then this will fail. Again, this is required if
          an id is not provided, and vice-versa.
        api_key (optional) - String: Your API key, required only if the key has not been declared
          elsewhere. This allows the API to recognize a request as yours and automatically route it
          to the appropriate destination.
        cloud (optional) - String: Your private cloud domain, required only if the key has not been declared
          elsewhere. This allows the API to recognize a request as yours and automatically route it
          to the appropriate destination.
        """
        batch = detect_batch(data)
        data = data_preprocess(data, batch=batch)
        url_params = {
            "batch": batch,
            "api_key": api_key,
            "version": version,
            "method": "remove_example",
        }
        return self._api_handler(
            data, cloud=cloud, api="custom", url_params=url_params, **kwargs
        )

    def wait(self, interval=1, timeout=None, **kwargs):
        """
        Block until the collection's model is completed training
        """
        start = time.time()
        while True:
            status = self.info(**kwargs).get("status")
            if timeout and time.time() - start > timeout:
                break
            if status == "ready":
                break
            if status != "training":
                raise IndicoError("Collection status failed with: {0}".format(status))
            time.sleep(interval)

    def register(
        self, make_public=False, cloud=None, api_key=None, version=None, **kwargs
    ):
        """
        This API endpoint allows you to register you collection in order to share read or write
        access to the collection with another user.

        Inputs:
        api_key (optional) - String: Your API key, required only if the key has not been declared
          elsewhere. This allows the API to recognize a request as yours and automatically route it
          to the appropriate destination.
        cloud (optional) - String: Your private cloud domain, required only if the key has not been declared
          elsewhere. This allows the API to recognize a request as yours and automatically route it
          to the appropriate destination.
        make_public (optional) - Boolean: When True, this option gives all indico users read access to your model.
        """
        kwargs["make_public"] = make_public
        url_params = {
            "batch": False,
            "api_key": api_key,
            "version": version,
            "method": "register",
        }
        return self._api_handler(
            None, cloud=cloud, api="custom", url_params=url_params, **kwargs
        )

    def deregister(self, cloud=None, api_key=None, version=None, **kwargs):
        """
        If you've shared access to your collection in the past, and now want to revoke all other user's access
        to your collection, simply deregister your collection.

        Inputs:
        api_key (optional) - String: Your API key, required only if the key has not been declared
          elsewhere. This allows the API to recognize a request as yours and automatically route it
          to the appropriate destination.
        cloud (optional) - String: Your private cloud domain, required only if the key has not been declared
          elsewhere. This allows the API to recognize a request as yours and automatically route it
          to the appropriate destination.
        """
        url_params = {
            "batch": False,
            "api_key": api_key,
            "version": version,
            "method": "deregister",
        }
        return self._api_handler(
            None, cloud=cloud, api="custom", url_params=url_params, **kwargs
        )

    def authorize(
        self,
        email,
        permission_type="read",
        cloud=None,
        api_key=None,
        version=None,
        **kwargs
    ):
        """
        This API endpoint allows you to authorize another user to access your model in a read or write capacity.
        Before calling authorize, you must first make sure your model has been registered.

        Inputs:
        email - String: The email of the user you would like to share access with.
        permission_type (optional) - String: One of ['read', 'write'].  Users with read permissions can only call `predict`.
          Users with `write` permissions can add new input examples and train models.
        api_key (optional) - String: Your API key, required only if the key has not been declared
          elsewhere. This allows the API to recognize a request as yours and automatically route it
          to the appropriate destination.
        cloud (optional) - String: Your private cloud domain, required only if the key has not been declared
          elsewhere. This allows the API to recognize a request as yours and automatically route it
          to the appropriate destination.
        """
        kwargs["permission_type"] = permission_type
        kwargs["email"] = email
        url_params = {
            "batch": False,
            "api_key": api_key,
            "version": version,
            "method": "authorize",
        }
        return self._api_handler(
            None, cloud=cloud, api="custom", url_params=url_params, **kwargs
        )

    def deauthorize(self, email, cloud=None, api_key=None, version=None, **kwargs):
        """
        This API endpoint allows you to remove another user's access to your collection.

        Inputs:
        email - String: The email of the user you would like to share access with.
        api_key (optional) - String: Your API key, required only if the key has not been declared
          elsewhere. This allows the API to recognize a request as yours and automatically route it
          to the appropriate destination.
        cloud (optional) - String: Your private cloud domain, required only if the key has not been declared
          elsewhere. This allows the API to recognize a request as yours and automatically route it
          to the appropriate destination.
        """
        kwargs["email"] = email
        url_params = {
            "batch": False,
            "api_key": api_key,
            "version": version,
            "method": "deauthorize",
        }
        return self._api_handler(
            None, cloud=cloud, api="custom", url_params=url_params, **kwargs
        )

    def rename(self, name, cloud=None, api_key=None, version=None, **kwargs):
        """
        If you'd like to change the name you use to access a given collection, you can call the rename endpoint.
        This is especially useful if the name you use for your model is not available for registration.

        Inputs:
        name - String: The new name used to access your model.
        api_key (optional) - String: Your API key, required only if the key has not been declared
          elsewhere. This allows the API to recognize a request as yours and automatically route it
          to the appropriate destination.
        cloud (optional) - String: Your private cloud domain, required only if the key has not been declared
          elsewhere. This allows the API to recognize a request as yours and automatically route it
          to the appropriate destination.
        """
        kwargs["name"] = name
        url_params = {
            "batch": False,
            "api_key": api_key,
            "version": version,
            "method": "rename",
        }
        result = self._api_handler(
            None, cloud=cloud, api="custom", url_params=url_params, **kwargs
        )
        self.keywords["collection"] = name
        return result


def collections(cloud=None, api_key=None, version=None, **kwargs):
    """
    This is a status report endpoint. It is used to get the status on all of the collections currently trained, as
    well as some basic statistics on their accuracies.

    Inputs
    api_key (optional) - String: Your API key, required only if the key has not been declared
      elsewhere. This allows the API to recognize a request as yours and automatically route it
      to the appropriate destination.
    cloud (optional) - String: Your private cloud domain, required only if the key has not been declared
      elsewhere. This allows the API to recognize a request as yours and automatically route it
      to the appropriate destination.

    Example usage:

      .. code-block:: python

         >>> collections = indicoio.collections()
        {
          "tag_predictor": {
            "input_type": "text",
            "model_type": "classification",
            "number_of_samples": 224
            'status': 'ready'
          }, "popularity_predictor": {
            "input_type": "text",
            "model_type": "regression",
            "number_of_samples": 231
            'status': 'training'
          }
        }
      }
    """
    url_params = {
        "batch": False,
        "api_key": api_key,
        "version": version,
        "method": "collections",
    }
    return api_handler(None, cloud=cloud, api="custom", url_params=url_params, **kwargs)


def vectorize(data, cloud=None, api_key=None, version=None, **kwargs):
    """
    Support for raw features from the custom collections API
    """
    batch = detect_batch(data)
    data = data_preprocess(data, batch=batch)
    url_params = {
        "batch": batch,
        "api_key": api_key,
        "version": version,
        "method": "vectorize",
    }
    return api_handler(data, cloud=cloud, api="custom", url_params=url_params, **kwargs)


class FinetuneCollection(Collection):
    def __init__(self, collection, **kwargs):
        self.keywords = {
            "shared": kwargs.get("shared"),
            "collection_name": collection,
            "job_options": {"job": True},
        }
        self.keywords.update(kwargs)

    def add_data(self, *args, **kwargs):
        raise NotImplementedError("Add Data not supported yet.")

    def train(self, *args, **kwargs):
        raise NotImplementedError("Train not supported yet.")

    def clear(self, *args, **kwargs):
        raise NotImplementedError("Clear not supported yet.")

    def explain(self, *args, **kwargs):
        raise NotImplementedError("Explain not supported yet.")

    def remove_example(self, *args, **kwargs):
        raise NotImplementedError("Remove Example not supported yet.")

    def rename(self, *args, **kwargs):
        raise NotImplementedError("Rename not supported yet.")

    def authorize(self, *args, **kwargs):
        return super().authorize(*args, **kwargs)

    def deauthorize(self, *args, **kwargs):
        return super().deauthorize(*args, **kwargs)

    def register(self, *args, **kwargs):
        return super().register(*args, **kwargs)

    def deregister(self, *args, **kwargs):
        return super().deregister(*args, **kwargs)

    def info(self, version=2, cloud=None, **kwargs):
        return self._api_handler(
            None, cloud=cloud, api="custom", method="info", version=version, **kwargs
        )

    def predict(self, data, version=2, cloud=None, **kwargs):
        return self._api_handler(
            data, cloud=cloud, api="custom", method="predict", version=version, **kwargs
        )

    def load(self, version=2, cloud=None, **kwargs):
        info = self.info(version=version, cloud=cloud, **kwargs)
        if info.get("load_status") == "ready":
            return info

        return self._api_handler(
            None, cloud=cloud, api="custom", method="load", version=version, **kwargs
        )



from ..utils.api import api_handler
from ..utils.decorators import detect_batch_decorator


@detect_batch_decorator
def sentiment_hq(text, cloud=None, batch=False, api_key=None, version=None, **kwargs):
    """
    Given input text, returns a scalar estimate of the sentiment of that text.
    Values are roughly in the range 0 to 1 with 0.5 indicating neutral sentiment.
    For reference, 0 suggests very negative sentiment and 1 suggests very positive sentiment.

    Example usage:

    .. code-block:: python

       >>> from indicoio import sentimenthq
       >>> text = 'Thanks everyone for the birthday wishes!! It was a crazy few days ><'
       >>> sentiment = sentimenthq(text)
       >>> sentiment
       0.6210052967071533

    :param text: The text to be analyzed.
    :type text: str or unicode
    :rtype: Float
    """
    url_params = {"batch": batch, "api_key": api_key, "version": version}
    return api_handler(
        text, cloud=cloud, api="sentimenthq", url_params=url_params, **kwargs
    )



from ..utils.api import api_handler
from ..utils.decorators import detect_batch_decorator


@detect_batch_decorator
def political(text, cloud=None, batch=False, api_key=None, version=2, **kwargs):
    """
    Given input text, returns a probability distribution over the political alignment of the speaker.

    Example usage:

    .. code-block:: python

       >>> from indicoio import political
       >>> import numpy as np
       >>> text = 'Wish we had more bike lanes. \
       Hopefully, driverless cars will chance economics from ownership to fee for service.'
       >>> affiliation = political(text)
       >>> affiliation
       {u'Libertarian': 0.4923755446986322, u'Green': 0.2974443102818122,
       u'Liberal': 0.13730032938784784, u'Conservative': 0.07287981563170784}
       >>> least_like = affiliation.keys()[np.argmin(affiliation.values())]
       >>> most_like = affiliation.keys()[np.argmax(affiliation.values())]
       >>> 'This text is most like %s and least like %s'%(most_like,least_like)
       u'This text is most like Libertarian and least like Conservative'

    :param text: The text to be analyzed.
    :type text: str or unicode
    :rtype: Dictionary of party probability pairs
    """
    url_params = {"batch": batch, "api_key": api_key, "version": version}
    return api_handler(
        text, cloud=cloud, api="political", url_params=url_params, **kwargs
    )



from base64 import b64encode, b64decode
import os


def docx_preprocess(docx, batch=False):
    """
    Load docx files from local filepath if not already b64 encoded
    """
    if batch:
        return [docx_preprocess(doc, batch=False) for doc in docx]

    if os.path.isfile(docx):
        # a filepath is provided, read and encode
        return b64encode(open(docx, "rb").read())
    else:
        # assume doc is already b64 encoded
        return docx



"""
Handles making requests to the IndicoApi Server
"""
from copy import deepcopy
import datetime
import json
import os.path
import sys
import time
import warnings

import requests
import msgpack
import msgpack_numpy as m
from indicoio import JSON_HEADERS, config
from indicoio.utils.encoder import NumpyEncoder
from indicoio.utils.errors import (
    APIDoesNotExist,
    BatchProcessingError,
    IndicoError,
    convert_to_py_error,
)

try:
    from urllib import urlencode
    from urlparse import urlparse
except Exception:  # For Python 3:
    from urllib.parse import urlencode, urlparse


m.patch()


class JobResult(object):
    def __init__(self, task_id, block=True, timeout=None, job=True):
        self.task_id = task_id
        self.block = block
        self.timeout = timeout

    def status(self, api_key=None, cloud=None):
        return api_handler(
            None,
            cloud=cloud,
            api="async",
            url_params={"method": "{task_id}/status".format(task_id=self.task_id)},
        )

    def get(self, api_key=None, cloud=None):
        if self.block:
            start_time = time.time()
            while True:
                status = self.status()
                if status in {"SUCCESS", "FAILURE", "REVOKED"}:
                    break

                if self.timeout and time.time() - start_time > self.timeout:
                    raise IndicoError(
                        "JobResult didn't finish in provided timeout {timeout}".format(
                            timeout=self.timeout
                        )
                    )
                time.sleep(2)

        return api_handler(
            None, cloud=cloud, api="async", url_params={"method": self.task_id}
        )


def convert(data):
    if isinstance(data, bytes):
        return data.decode("utf-8", "ignore")
    if isinstance(data, dict):
        return dict(map(convert, data.items()))
    if isinstance(data, tuple):
        return map(convert, data)
    if isinstance(data, list):
        return list(map(convert, data))
    return data


def batched(iterable, size):
    """
    Split an iterable into constant sized chunks
    Recipe from http://stackoverflow.com/a/8290514
    """
    length = len(iterable)
    for batch_start in range(0, length, size):
        yield iterable[batch_start : batch_start + size]


def standardize_input_data(data):
    """
    Ensure utf-8 encoded strings are passed to the indico API
    """
    if type(data) == bytes:
        data = data.decode("utf-8")
    if type(data) == list:
        data = [standardize_input_data(el) for el in data]
    return data


def api_handler(
    input_data, cloud, api, url_params=None, batch_size=None, job_options=None, **kwargs
):
    """
    Sends finalized request data to ML server and receives response.
    If a batch_size is specified, breaks down a request into smaller
    component requests and aggregates the results.
    """
    job_options = deepcopy(job_options or {})
    kwargs["job"] = job = job_options.pop("job", False)

    url_params = url_params or {}
    input_data = standardize_input_data(input_data)

    cloud = cloud or config.cloud
    host = "%s.indico.domains" % cloud if cloud else config.host

    # LOCAL DEPLOYMENTS
    url_protocol = config.url_protocol

    headers = dict(JSON_HEADERS)
    headers["X-ApiKey"] = url_params.get("api_key") or config.api_key
    url = create_url(url_protocol, host, api, dict(kwargs, **url_params))
    return collect_api_results(
        input_data,
        url,
        headers,
        api,
        batch_size,
        kwargs,
        job=job,
        job_options=job_options,
    )


def collect_api_results(
    input_data, url, headers, api, batch_size, kwargs, job=False, job_options=None
):
    """
    Optionally split up a single request into a series of requests
    to ensure timely HTTP responses.

    Could eventually speed up the time required to receive a response by
    sending batches to the indico API concurrently
    """
    if batch_size:
        results = []
        for batch in batched(input_data, size=batch_size):
            try:
                result = send_request(batch, api, url, headers, kwargs)
                if isinstance(result, list):
                    results.extend(result)
                else:
                    results.append(result)
            except IndicoError as e:
                # Log results so far to file
                timestamp = datetime.datetime.now().strftime("%Y-%m-%d-%H:%M:%S")
                filename = "indico-{api}-{timestamp}.json".format(
                    api=api, timestamp=timestamp
                )
                if sys.version_info > (3, 0):
                    json.dump(
                        results,
                        open(filename, mode="w", encoding="utf-8"),
                        cls=NumpyEncoder,
                    )
                else:
                    json.dump(results, open(filename, mode="w"), cls=NumpyEncoder)
                raise BatchProcessingError(
                    "The following error occurred while processing your data: `{err}` "
                    "Partial results have been saved to {filename}".format(
                        err=e, filename=os.path.abspath(filename)
                    )
                )
        if job:
            results = [JobResult(result, **job_options) for result in results]
            results = [result.get() for result in results]
        return results

    else:
        result = send_request(input_data, api, url, headers, kwargs)
        if job:
            result = JobResult(result, **job_options).get()
        return result


def send_request(input_data, api, url, headers, kwargs):
    """
    Use the requests library to send of an HTTP call to the indico servers
    """
    data = {}
    if input_data != None:
        data["data"] = input_data

    # request that the API respond with a msgpack encoded result
    serializer = kwargs.pop("serializer", config.serializer)
    data["serializer"] = serializer

    data.update(**kwargs)

    json_data = json.dumps(data)

    response = requests.post(url, data=json_data, headers=headers)
    warning = response.headers.get("x-warning")
    if warning:
        warnings.warn(warning)

    cloud = urlparse(url).hostname
    if response.status_code == 503 and not cloud.endswith(".indico.io"):
        raise APIDoesNotExist(
            "Private cloud '%s' does not include api '%s'" % (cloud, api)
        )

    try:
        if serializer == "msgpack":
            json_results = msgpack.unpackb(response.content)
        else:
            json_results = response.json()
    except (msgpack.exceptions.UnpackException, msgpack.exceptions.ExtraData):
        try:
            json_results = response.json()
        except Exception:
            json_results = {"error": response.text}

    if config.PY3:
        json_results = convert(json_results)

    results = json_results.get("results", False)
    if results is False:
        error = json_results.get("error")
        raise convert_to_py_error(error)
    return results


def create_url(url_protocol, host, api, url_params):
    """
    Generate the proper url for sending off data for analysis
    """
    is_batch = url_params.pop("batch", None)
    apis = url_params.pop("apis", None)
    version = url_params.pop("version", None) or url_params.pop("v", None)
    method = url_params.pop("method", None)

    host_url_seg = url_protocol + "://%s" % host
    api_url_seg = "/%s" % api
    batch_url_seg = "/batch" if is_batch else ""
    method_url_seg = "/%s" % method if method else ""

    params = {}
    if apis:
        params["apis"] = ",".join(apis)
    if version:
        params["version"] = version

    url = host_url_seg + api_url_seg + batch_url_seg + method_url_seg
    if params:
        url += "?" + urlencode(params)

    return url




from ..utils.api import api_handler
from ..utils.decorators import detect_batch_decorator


@detect_batch_decorator
def text_tags(text, cloud=None, batch=False, api_key=None, version=None, **kwargs):
    """
    Given input text, returns a probability distribution over 100 document categories

    Example usage:

    .. code-block:: python

       >>> import indicoio
       >>> import numpy as np
       >>> text = 'Monday: Delightful with mostly sunny skies. Highs in the low 70s.'
       >>> possible = indicoio.classification(text)
       >>> category = possible.keys()[np.argmax(possible.values())]
       >>> probability = np.max(possible.values())
       >>> "Predicted category '%s' with probability %.4f"%(category,probability)
       u'Predicted 'Weather' with probability 0.8548'

    :param text: The text to be analyzed.
    :type text: str or unicode
    :rtype: Dictionary of class probability pairs
    """
    url_params = {"batch": batch, "api_key": api_key, "version": version}
    return api_handler(
        text, cloud=cloud, api="texttags", url_params=url_params, **kwargs
    )



from functools import wraps, partial
import warnings


Version, version, __version__, VERSION = ("1.3.4",) * 4

JSON_HEADERS = {
    "Content-type": "application/json",
    "Accept": "application/json",
    "client-lib": "python",
    "version-number": VERSION,
}

from .text import *
from .image import *
from .multi import *
from .pdf import *
from .docx import *

from indicoio.utils.errors import (
    IndicoError,
    InvalidAPIKey,
    MissingAPIKey,
    MalformattedData,
    BatchProcessingError,
    APIDoesNotExist,
)



from ..utils.preprocessing import data_preprocess
from ..utils.api import api_handler
from ..utils.decorators import detect_batch_decorator


@detect_batch_decorator
def image_recognition(
    image, cloud=None, batch=False, api_key=None, version=None, **kwargs
):
    """
    Given an input image, returns a dictionary of image classifications with associated scores

    * Input can be either grayscale or rgb color and should either be a numpy array or nested list format.
    * Input data should be either uint8 0-255 range values or floating point between 0 and 1.
    * Large images (i.e. 1024x768+) are much bigger than needed, minaxis resizing will be done internally to 144 if needed.
    * For ideal performance, images should be square aspect ratio but non-square aspect ratios are supported as well.

    Example usage:

    .. code-block:: python

       >>> from indicoio import image_recognition
       >>> features = image_recognition(<filename>)

    :param image: The image to be analyzed.
    :type image: str
    :rtype: dict containing classifications
    """
    image = data_preprocess(image, batch=batch, size=144, min_axis=True)
    url_params = {"batch": batch, "api_key": api_key, "version": version}
    return api_handler(
        image, cloud=cloud, api="imagerecognition", url_params=url_params, **kwargs
    )



from ..utils.api import api_handler
from ..utils.preprocessing import data_preprocess
from ..utils.decorators import detect_batch_decorator


@detect_batch_decorator
def content_filtering(
    image, cloud=None, batch=False, api_key=None, version=None, **kwargs
):
    """
    Given a grayscale input image, returns how obcene the image is.
    Input should be in a list of list format.

    Example usage:

    .. code-block:: python

       >>> from indicoio import content_filtering
       >>> import numpy as np
       >>> face = np.zeros((48,48)).tolist()
       >>> res = content_filtering(face)
       >>> res
	   .056

    :param image: The image to be analyzed.
    :type image: list of lists
    :rtype: float of nsfwness
    """
    image = data_preprocess(image, batch=batch, size=128, min_axis=True)
    url_params = {"batch": batch, "api_key": api_key, "version": version}
    return api_handler(
        image, cloud=cloud, api="contentfiltering", url_params=url_params, **kwargs
    )



"""
Contains Indico Custom Errors
"""


# skeleton error classes to allow
# more specific exception handling
class IndicoError(ValueError):
    pass


class InvalidAPIKey(IndicoError):
    pass


class MissingAPIKey(IndicoError):
    pass


class MalformattedData(IndicoError):
    pass


class APIDoesNotExist(IndicoError):
    pass


class BatchProcessingError(IndicoError):
    pass


class DataStructureException(Exception):
    """
    If a non-accepted datastructure is passed, throws an exception
    """

    def __init__(self, callback, passed_structure, accepted_structures):
        self.callback = callback.__name__
        self.structure = str(type(passed_structure))
        self.accepted = [str(structure) for structure in accepted_structures]

    def __str__(self):
        return """
        function %s does not accept %s, accepted types are: %s
        """ % (
            self.callback,
            self.structure,
            str(self.accepted),
        )


ERR_MSGS = (
    ("invalid api key", InvalidAPIKey),
    ("api key not found", MissingAPIKey),
    ("input contains one or more empty strings", MalformattedData),
)


def convert_to_py_error(error_message):
    """
    Raise specific exceptions for ease of error handling
    """
    message = error_message.lower()
    for err_msg, err_type in ERR_MSGS:
        if err_msg in message:
            return err_type(error_message)
    else:
        return IndicoError(error_message)



from ..utils.api import api_handler
from ..utils.decorators import detect_batch_decorator


@detect_batch_decorator
def emotion(text, cloud=None, batch=False, api_key=None, version=None, **kwargs):
    """
    Given input text, returns a probability distribution over 5 possible
    emotions of what language the text was written in.

    Example usage:

    .. code-block:: python

       >>> import indicoio
       >>> import numpy as np
       >>> text = "I did it. I got into Grad School. Not just any program, but a GREAT program. :-)"
       >>> possible = indicoio.emotion(text)
       >>> emotion = possible.keys()[np.argmax(possible.values())]
       >>> probability = np.max(possible.values())
       >>> 'Predicted `%s` with probability %.4f' % (emotion, probability)
       u'Predicted `joy` with probability 0.7744'

    :param text: The text to be analyzed.
    :type text: str or unicode
    :rtype: Dictionary of emotion probability pairs
    """
    url_params = {"batch": batch, "api_key": api_key, "version": version}
    return api_handler(
        text, cloud=cloud, api="emotion", url_params=url_params, **kwargs
    )



"""
Setup for indico apis
"""
import sys

PY2 = True if sys.version_info[0] == 2 else False
PY3 = not PY2

REQUIREMENTS = [
    "msgpack==0.5.6",
    "msgpack-numpy==0.4.1",
    "Pillow<6.0.0",
    "requests>=2.2.1",
    "setuptools==36.5.0",
    "six>=1.3.0",
]

# shim for futures support
if PY2:
    REQUIREMENTS.append("futures>=3.0.0")
    REQUIREMENTS.append("numpy<1.17")
else:
    REQUIREMENTS.append("numpy>=1.14.1")

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


setup(
    name="IndicoIo",
    version="1.3.4",
    packages=[
        "indicoio",
        "indicoio.text",
        "indicoio.image",
        "indicoio.multi",
        "indicoio.utils",
        "indicoio.custom",
        "indicoio.pdf",
        "indicoio.docx",
        "tests",
    ],
    description="""A Python Wrapper for indico. Use pre-built state of the art machine learning algorithms with a single line of code.""",
    license="MIT License (See LICENSE)",
    long_description=open("README.rst").read(),
    url="https://github.com/IndicoDataSolutions/indicoio-python",
    author="Alec Radford, Slater Victoroff, Aidan McLaughlin, Madison May, Anne Carlson",
    author_email="engineering@indico.io",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Image Recognition",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    setup_requires=["six >= 1.3.0", "pillow >= 2.8.1"],
    tests_require=["mock >= 1.3.0, < 2.0.0", "nose >= 1.0", "numpy == 1.16.2"],
    test_suite="nose.collector",
    install_requires=REQUIREMENTS,
)



from ..utils.pdf import pdf_preprocess, postprocess_images
from ..utils.api import api_handler
from ..utils.decorators import detect_batch_decorator


@detect_batch_decorator
def pdf_extraction(pdf, cloud=None, batch=False, api_key=None, version=None, **kwargs):
    """
    Given a pdf, returns the text and metadata associated with the given pdf.
    PDFs may be provided as base64 encoded data or as a filepath.
    Base64 image data and formatted table is optionally returned by setting
    `images=True` or `tables=True`.

    Example usage:

    .. code-block:: python

       >>> from indicoio import pdf_extraction
       >>> results = pdf_extraction(pdf_file)
       >>> results.keys()
       ['text', 'metadata']

    :param pdf: The pdf to be analyzed.
    :type pdf: str or list of strs
    :rtype: dict or list of dicts
    """
    pdf = pdf_preprocess(pdf, batch=batch)
    url_params = {"batch": batch, "api_key": api_key, "version": version}

    if version == 2:
        kwargs["job_options"] = kwargs.get("job_options", {"job": True})
        kwargs["job_options"]["job"] = True

    results = api_handler(
        pdf, cloud=cloud, api="pdfextraction", url_params=url_params, **kwargs
    )

    if batch:
        for result in results:
            result["images"] = postprocess_images(result.get("images", []))
    else:
        results["images"] = postprocess_images(results.get("images", []))
    return results



from ..utils.preprocessing import data_preprocess
from ..utils.api import api_handler
from ..utils.decorators import detect_batch_decorator


@detect_batch_decorator
def image_features(image, cloud=None, batch=False, api_key=None, version=3, **kwargs):
    """
    Given an input image, returns a 2048 dimensional sparse feature vector explaining that image.
    Useful as a form of feature engineering for image oriented tasks.

    * Input can be either grayscale or rgb color and should either be a numpy array or nested list format.
    * Input data should be either uint8 0-255 range values or floating point between 0 and 1.
    * Large images (i.e. 1024x768+) are much bigger than needed, resizing will be done internally to 144x144 if needed.
    * For ideal performance, images should be square aspect ratio but non-square aspect ratios are supported as well.

    Example usage:

    .. code-block:: python

       >>> from indicoio import image_features
       >>> import numpy as np
       >>> image = np.zeros((144,144,3))
       >>> features = image_features(image)
       >>> len(features),np.min(features),np.max(features),np.sum(np.asarray(f)!=0)
       (2048, 0.0, 6.97088623046875, 571)

    Since the image features returned are a semantic description of the contents of an image they can be used
    to implement many other common image related tasks such as object recognition or image similarity and retrieval.

    For image similarity, simple distance metrics applied to collections of image feature vectors can work very well.

    :param image: The image to be analyzed.
    :type image: numpy.ndarray
    :rtype: List containing features
    """
    image = data_preprocess(image, batch=batch, size=512, min_axis=True)
    url_params = {"batch": batch, "api_key": api_key, "version": version}
    return api_handler(
        image, cloud=cloud, api="imagefeatures", url_params=url_params, **kwargs
    )
